class Config(object):
    @property
    def list_ignore_case(self):
        """
        Ignore case while doing `IN_LIST` operation
        """
        return True

    @property
    def implicit_punct(self):
        """
        Automatically add optional Punctuation characters inside rule between macros.
        eg. `WORD(w1), WORD(w2)`
        would be converted into:
        `WORD(w1), PUNCT?, WORD(w2)`
        """
        return True



import json
import argparse

import rita


def main():
    parser = argparse.ArgumentParser(description="Compile rita -> spaCy patterns")

    parser.add_argument("-f", help=".rita rules file")
    parser.add_argument("out", help="output .jsonl file to store rules")
    args = parser.parse_args()

    patterns = rita.compile(args.f)

    with open(args.out, "w") as f:
        for pattern in patterns:
            f.write(json.dumps(pattern) + "\n")



import logging

from functools import partial

logger = logging.getLogger(__name__)


def any_of_parse(lst, op=None):
    extra = []

    def is_simple(w):
        return not "-" in w

    # Covers basic case when all words are normal
    normalized = set(map(lambda x: x.lower(), lst))
    simple = set(filter(is_simple, normalized))
    other = normalized - simple
    
    base = {"LOWER": {"REGEX": r"({0})".format("|".join(simple))}}
    if op:
        base["OP"] = op

    if len(other) > 0:
        base["OP"] = "?"
    yield base

    # Not very good solution
    # This should be separate rules instead of single one with bunch of `?` operators

    for item in other:
        buff = item.split("-")
        yield next(generic_parse("ORTH", buff[0], "?"))
        for b in buff[1:]:
            yield next(generic_parse("ORTH", "-", "?"))
            yield next(generic_parse("ORTH", b, "?"))


def regex_parse(r, op=None):
    d = {"TEXT": {"REGEX": r}}

    if op:
        d["OP"] = op
    yield d


def fuzzy_parse(r, op=None):
    # TODO: build premutations
    d = {"LOWER": {"REGEX": "({0})[.,?;!]?".format("|".join(r))}}
    if op:
        d["OP"] = op
    yield d


def generic_parse(tag, value, op=None):
    d = {}
    d[tag] = value
    if op:
        d["OP"] = op
    yield d

def punct_parse(_, op=None):
    d = {}
    d["IS_PUNCT"] = True
    if op:
        d["OP"] = op
    yield d

PARSERS = {
    "any_of": any_of_parse,
    "value": partial(generic_parse, "ORTH"),
    "regex": regex_parse,
    "entity": partial(generic_parse, "ENT_TYPE"),
    "lemma": partial(generic_parse, "LEMMA"),
    "pos": partial(generic_parse, "POS"),
    "punct": punct_parse,
    "fuzzy": fuzzy_parse,
}


def rules_to_patterns(label, data):
    return {
        "label": label,
        "pattern": [p
                    for (t, d, op) in data
                    for p in PARSERS[t](d, op)],
    }


def compile_rules(rules):
    logger.info("Using spaCy rules implementation")
    return [rules_to_patterns(*group)
            for group in rules]



import logging
import types

from rita import engine
from rita.config import Config
from rita.parser import RitaParser
from rita.preprocess import preprocess_rules


logger = logging.getLogger(__name__)


def config():
    return Config()


def compile_string(raw, compile_fn=None):
    parser = RitaParser()
    parser.build()
    root = parser.parse(raw)
    logger.debug(root)
    if compile_fn:
        compile_rules = compile_fn
    else:
        compile_rules = engine.get_default()

    rules = list(preprocess_rules(root))
    result = compile_rules(rules)
    if isinstance(result, types.GeneratorType):
        return list(result)
    else:
        return result


def compile(fname, compile_fn=None):
    with open(fname, "r") as f:
        raw = f.read()

    return compile_string(raw, compile_fn)



import logging

import rita

from functools import reduce
from itertools import zip_longest

logger = logging.getLogger(__name__)


def add_implicit_punct(rules):
    """
    When writing rule, user usually doesn't care about some punct characters between words.
    We add them implicitly (unless this setting is turned off)
    """
    for group_label, pattern in rules:
        def gen():
            for p in pattern:
                yield p
                yield ("punct", None, "?")

        if len(pattern) == 1:
            yield (group_label, pattern)
        else:
            yield (group_label, list(gen())[:-1])


def handle_multi_word(rules):
    """
    spaCy splits everything in tokens. Words with dash ends up in different tokens.
    We don't want for user to even care about this, so we make this work implicitly
    
    WORD("Knee-length") => WORD("Knee"), WORD("-"), WORD("length")
    """
    for group_label, pattern in rules:
        yield (group_label, pattern)


def dummy(rules):
    """
    Placeholder which does nothing
    """
    return rules


def rule_tuple(d):
    return (d["label"], d["data"])


def preprocess_rules(root):
    logger.info("Preprocessing rules")
    conf = rita.config()

    rules = [rule_tuple(doc())
             for doc in root
             if doc and doc()]

    pipeline = [dummy, handle_multi_word]

    if conf.implicit_punct:
        logger.info("Adding implicit Punctuations")
        pipeline.append(add_implicit_punct)

    return reduce(lambda acc, p: p(acc), pipeline, rules)
